import time

from jnt_django_toolbox.profiling.profilers.base import BaseProfiler


class RequestsProfiler(BaseProfiler):
    """Request profiler."""

    def __init__(self, header: str = "Request-Stats"):
        """Initializing."""
        self._header = header

    def before_request(self, request, stack):
        """Start capturing requests."""
        self._start_time = time.time()

    def after_request(self, request, response):
        """Add profiling info to response."""
        response[self._header] = "req_time={0:.3f}".format(
            time.time() - self._start_time,
        )



from contextlib import suppress

from jnt_django_toolbox.models.fields.bit.types import Bit, BitHandler


class BitQueryLookupWrapper:
    """Bit query lookup wrapper."""

    def __init__(self, alias, column, bit):
        """Initializing."""
        self.table_alias = alias
        self.column = column
        self.bit = bit

    def as_sql(self, qn, connection=None):
        """
        Create the proper SQL fragment.

        This inserts something like "(T0.flags & value) != 0". This will be
        called by Where.as_sql()
        """
        if self.bit:
            return (
                "(%s.%s | %d)"
                % (qn(self.table_alias), qn(self.column), self.bit.mask),
                [],
            )
        return (
            "(%s.%s & %d)"
            % (qn(self.table_alias), qn(self.column), self.bit.mask),
            [],
        )


with suppress(ImportError):
    from django.db.models.lookups import Exact

    class BitQueryLookupWrapper(Exact):  # NOQA
        def process_lhs(self, qn, connection, lhs=None):
            """Process left."""
            lhs_sql, params = super().process_lhs(qn, connection, lhs)
            if self.rhs:
                lhs_sql = "{0} & %s".format(lhs_sql)
            else:
                lhs_sql = "{0} | %s".format(lhs_sql)
            params.extend(self.get_db_prep_lookup(self.rhs, connection)[1])
            return lhs_sql, params

        def get_db_prep_lookup(self, value, connection, prepared=False):
            """Get db prepared lookup."""
            v = value.mask if isinstance(value, (BitHandler, Bit)) else value
            return super().get_db_prep_lookup(v, connection)

        def get_prep_lookup(self):
            """Get prepared lookup."""
            return self.rhs


class BitQuerySaveWrapper(BitQueryLookupWrapper):
    """Bit query save wrapper."""

    def as_sql(self, qn, connection):
        """
        Create the proper SQL fragment.

        This inserts something like "(T0.flags & value) != 0". This will be
        called by Where.as_sql()
        """
        engine = connection.settings_dict["ENGINE"].rsplit(".", -1)[-1]
        if engine.startswith("postgres"):
            XOR_OPERATOR = "#"
        elif engine.startswith("sqlite"):
            raise NotImplementedError
        else:
            XOR_OPERATOR = "^"

        if self.bit:
            return (
                "%s.%s | %d"
                % (qn(self.table_alias), qn(self.column), self.bit.mask),
                [],
            )
        return (
            "%s.%s %s %d"
            % (
                qn(self.table_alias),
                qn(self.column),
                XOR_OPERATOR,
                self.bit.mask,
            ),
            [],
        )



from contextlib import suppress

import six
from django.db.models.fields import BigIntegerField, Field

from jnt_django_toolbox.admin.fields import BitFieldFormField
from jnt_django_toolbox.models.fields.bit.query import BitQueryLookupWrapper
from jnt_django_toolbox.models.fields.bit.types import Bit, BitHandler

# Count binary capacity. Truncate "0b" prefix from binary form.
# Twice faster than bin(i)[2:] or math.floor(math.log(i))
MAX_FLAG_COUNT = int(len(bin(BigIntegerField.MAX_BIGINT)) - 2)


class BitFieldFlags:
    """Bit field flags wrapper."""

    def __init__(self, flags):
        """Initializing."""
        if len(flags) > MAX_FLAG_COUNT:
            raise ValueError("Too many flags")

        self._flags = flags

    def items(self):
        """Items."""
        return [(flag, Bit(self._flags.index(flag))) for flag in self._flags]

    def keys(self):
        """Items."""
        return self._flags

    def values(self):
        """Values."""
        return [Bit(self._flags.index(flag)) for flag in self._flags]

    def __getattr__(self, key):
        if key not in self._flags:
            raise AttributeError("flag {0} is not registered".format(key))

        return Bit(self._flags.index(key))

    def __repr__(self):
        return repr(self._flags)

    def __iter__(self):
        yield from self._flags


class BitFieldCreator:
    """
    A placeholder class that provides a way to set the attribute on the model.

    Descriptor for BitFields.  Checks to make sure that all flags of the
    instance match the class.  This is to handle the case when caching
    an older version of the instance and a newer version of the class is
    available (usually during deploys).
    """

    def __init__(self, field):
        """Initializing."""
        self.field = field

    def __set__(self, instance, value):
        instance.__dict__[self.field.name] = self.field.to_python(value)

    def __get__(self, instance, type=None):
        if instance is None:
            return BitFieldFlags(self.field.flags)
        retval = instance.__dict__[self.field.name]
        if self.field.__class__ is BitField:
            # Update flags from class in case they've changed.
            retval._keys = self.field.flags
        return retval


class BitField(BigIntegerField):
    """Bit model field."""

    def __init__(self, flags, default=None, *args, **kwargs):
        """Initializing."""
        if isinstance(flags, dict):
            # Get only integer keys in correct range
            valid_keys = [
                flag_key
                for flag_key in flags.keys()
                if isinstance(flag_key, int)
                and (0 <= flag_key < MAX_FLAG_COUNT)
            ]
            if not valid_keys:
                raise ValueError("Wrong keys or empty dictionary")
            # Fill list with values from dict or with empty values
            flags = [flags.get(i, "") for i in range(max(valid_keys) + 1)]

        if len(flags) > MAX_FLAG_COUNT:
            raise ValueError("Too many flags")

        self._arg_flags = flags
        flags = list(flags)
        labels = []
        for num, flag in enumerate(flags):
            if isinstance(flag, (tuple, list)):
                flags[num] = flag[0]
                labels.append(flag[1])
            else:
                labels.append(flag)

        if isinstance(default, (list, tuple, set, frozenset)):
            new_value = 0
            for flag in default:
                new_value |= Bit(flags.index(flag))
            default = new_value

        BigIntegerField.__init__(self, default=default, *args, **kwargs)
        self.flags = flags
        self.labels = labels

    def contribute_to_class(self, cls, name, **kwargs):  # noqa: WPS117
        """Contribute to class."""
        super().contribute_to_class(cls, name, **kwargs)
        setattr(cls, self.name, BitFieldCreator(self))

    def formfield(self, form_class=BitFieldFormField, **kwargs):
        """Get form field."""
        choices = [(k, self.labels[self.flags.index(k)]) for k in self.flags]
        return Field.formfield(self, form_class, choices=choices, **kwargs)

    def pre_save(self, instance, add):
        """Before save."""
        return getattr(instance, self.attname)

    def get_prep_value(self, value):
        """Get prepared value."""
        if value is None:
            return None
        if isinstance(value, (BitHandler, Bit)):
            value = value.mask
        return int(value)

    def get_db_prep_lookup(
        self,
        lookup_type,
        value,
        connection,
        prepared=False,
    ):
        """Get db prepared lookup."""
        if isinstance(getattr(value, "expression", None), Bit):
            value = value.expression
        if isinstance(value, (BitHandler, Bit)):
            return [value.mask]
        return BigIntegerField.get_db_prep_lookup(
            self,
            lookup_type=lookup_type,
            value=value,
            connection=connection,
            prepared=prepared,
        )

    def get_prep_lookup(self, lookup_type, value):
        """Get prepared lookup."""
        if isinstance(getattr(value, "expression", None), Bit):
            value = value.expression
        if isinstance(value, Bit):
            if lookup_type == "exact":
                return value
            raise TypeError(
                "Lookup type {0!r} not supported with `Bit` type.".format(
                    lookup_type,
                ),
            )
        return BigIntegerField.get_prep_lookup(self, lookup_type, value)

    def to_python(self, value):
        """Provide raw value."""
        if isinstance(value, Bit):
            value = value.mask
        if not isinstance(value, BitHandler):
            # Regression for #1425: fix bad data that was created resulting
            # in negative values for flags.  Compute the value that would
            # have been visible ot the application to preserve compatibility.
            if isinstance(value, six.integer_types) and value < 0:
                new_value = 0
                for bit_number, _ in enumerate(self.flags):
                    new_value |= value & (2 ** bit_number)
                value = new_value

            value = BitHandler(value, self.flags, self.labels)
        else:
            # Ensure flags are consistent for unpickling
            value._keys = self.flags
        return value

    def deconstruct(self):
        """Deconstruct."""
        name, path, args, kwargs = super().deconstruct()
        args.insert(0, self._arg_flags)
        return name, path, args, kwargs


with suppress(AttributeError):
    BitField.register_lookup(BitQueryLookupWrapper)



from abc import ABC, abstractmethod
from contextlib import ExitStack

from django.http import HttpRequest, HttpResponse


class BaseProfiler(ABC):
    """Base class for custom profilers."""

    @abstractmethod
    def before_request(self, request: HttpRequest, stack: ExitStack):
        """Call on recieve request."""

    @abstractmethod
    def after_request(self, request: HttpRequest, response: HttpResponse):
        """Call after is processed and response available."""



import datetime
import json
from contextlib import suppress
from math import floor
from time import time

from django.template.defaultfilters import truncatechars
from django.utils.encoding import force_str
from opentracing import global_tracer


def wrap_cursor(connection):
    """Wrap cursor for logging sql queries into jaeger."""
    if not hasattr(connection, "_traced_cursor"):  # noqa: WPS421
        connection._traced_cursor = connection.cursor
        connection._traced_chunked_cursor = connection.chunked_cursor

        def cursor(*args, **kwargs):  # noqa: WPS430
            # Per the DB API cursor() does not accept any arguments. There"s
            # some code in the wild which does not follow that convention,
            # so we pass on the arguments even though it"s not clean.
            # See:
            # https://github.com/jazzband/django-debug-toolbar/pull/615
            # https://github.com/jazzband/django-debug-toolbar/pull/896
            return NormalCursorWrapper(
                connection._traced_cursor(*args, **kwargs),
                connection,
            )

        def chunked_cursor(*args, **kwargs):  # noqa: WPS430
            return NormalCursorWrapper(
                connection._traced_chunked_cursor(*args, **kwargs),
                connection,
            )

        connection.cursor = cursor
        connection.chunked_cursor = chunked_cursor
        return cursor


def unwrap_cursor(connection):
    """Unwrap cursor."""
    if hasattr(connection, "_traced_cursor"):  # noqa: WPS421
        del connection._traced_cursor  # noqa: WPS420
        del connection.cursor  # noqa: WPS420
        del connection.chunked_cursor  # noqa: WPS420


class NormalCursorWrapper:
    """Wraps a cursor and logs queries."""

    def __init__(self, cursor, db):
        """Initialize."""
        self.cursor = cursor
        # Instance of a BaseDatabaseWrapper subclass
        self.db = db

    def callproc(self, procname, params=None):
        """Wrap callproc."""
        return self._record(self.cursor.callproc, procname, params)

    def execute(self, sql, params=None):
        """Wrap execute."""
        return self._record(self.cursor.execute, sql, params)

    def executemany(self, sql, param_list):
        """Wrap executemany."""
        return self._record(self.cursor.executemany, sql, param_list)

    def __exit__(self, type_, value, traceback):
        self.close()

    def __getattr__(self, attr):
        return getattr(self.cursor, attr)

    def __iter__(self):
        return iter(self.cursor)

    def __enter__(self):
        return self

    def _quote_expr(self, element):
        if isinstance(element, str):
            return "'{0}'".format(element.replace("'", "''"))

        return repr(element)

    def _quote_params(self, params):
        if not params:
            return params
        if isinstance(params, dict):
            return {
                key: self._quote_expr(value) for key, value in params.items()
            }
        return [self._quote_expr(p) for p in params]

    def _decode(self, param):
        # If a sequence type, decode each element separately
        if isinstance(param, (tuple, list)):
            return [self._decode(element) for element in param]

        # If a dictionary type, decode each value separately
        if isinstance(param, dict):
            return {key: self._decode(value) for key, value in param.items()}

        # make sure datetime, date and time are converted to string by force_str
        convert_types = (datetime.datetime, datetime.date, datetime.time)
        try:
            return force_str(
                param,
                strings_only=not isinstance(param, convert_types),
            )
        except UnicodeDecodeError:
            return "(encoded string)"

    def _record(self, method, sql, params):
        start_time = time()

        # Sql might be an object (such as psycopg Composed).
        # For logging purposes, make sure it"s str.
        sql_str = str(sql)

        with global_tracer().start_active_span(
            truncatechars(sql_str, 90),
            start_time=start_time,
            tags={"scope": "sql"},
            child_of=global_tracer().active_span,
        ) as scope:

            try:  # noqa:WPS501
                return method(sql, params)

            finally:
                stop_time = time()
                duration = (stop_time - start_time) * 1000
                serialized_params = ""

                with suppress(TypeError):
                    serialized_params = json.dumps(self._decode(params))

                alias = getattr(self.db, "alias", "default")
                conn = self.db.connection
                vendor = getattr(conn, "vendor", "unknown")

                scope.span.log_kv(
                    {
                        "vendor": vendor,
                        "alias": alias,
                        "sql": self.db.ops.last_executed_query(
                            self.cursor,
                            sql_str,
                            self._quote_params(params),
                        ),
                        "duration": "{0} ms".format(floor(duration)),
                        "raw_sql": sql_str,
                        "params": serialized_params,
                        "raw_params": params,
                        "start_time": start_time,
                        "stop_time": stop_time,
                        "is_select": sql_str.lower()
                        .strip()
                        .startswith("select"),
                        "encoding": conn.encoding,
                        "trans_status": conn.get_transaction_status(),
                    },
                )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['jnt_django_toolbox',
 'jnt_django_toolbox.admin',
 'jnt_django_toolbox.admin.decorators',
 'jnt_django_toolbox.admin.fields',
 'jnt_django_toolbox.admin.filters',
 'jnt_django_toolbox.admin.helpers',
 'jnt_django_toolbox.admin.widgets',
 'jnt_django_toolbox.consts',
 'jnt_django_toolbox.context_managers',
 'jnt_django_toolbox.forms',
 'jnt_django_toolbox.forms.fields',
 'jnt_django_toolbox.helpers',
 'jnt_django_toolbox.management',
 'jnt_django_toolbox.management.commands',
 'jnt_django_toolbox.models',
 'jnt_django_toolbox.models.fields',
 'jnt_django_toolbox.models.fields.bit',
 'jnt_django_toolbox.profiling',
 'jnt_django_toolbox.profiling.db',
 'jnt_django_toolbox.profiling.decorators',
 'jnt_django_toolbox.profiling.profilers']

package_data = \
{'': ['*']}

install_requires = \
['dateparser', 'django>=3']

extras_require = \
{'jaeger': ['jaeger-client']}

setup_kwargs = {
    'name': 'jnt-django-toolbox',
    'version': '0.1.38',
    'description': '',
    'long_description': None,
    'author': 'Junte',
    'author_email': 'tech@junte.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from statistics import mean

from django.db import connection
from django.test.utils import CaptureQueriesContext

from jnt_django_toolbox.profiling.profilers.base import BaseProfiler


class DatabaseQueriesProfiler(BaseProfiler):
    """Database queries count profiler."""

    def __init__(self, header: str = "Db-Stats"):
        """Initializing."""
        self._header = header

    def before_request(self, request, stack):
        """Start capturing requests."""
        self._context = CaptureQueriesContext(connection)
        stack.enter_context(self._context)

    def after_request(self, request, response):
        """Add profiling info to response."""
        query_timings = [float(query["time"]) for query in self._context]
        stats = ["db_count={0}".format(len(self._context))]
        if query_timings:
            stats.extend(
                [
                    "db_total={0:.3f}".format(sum(query_timings)),
                    "db_max={0:.3f}".format(max(query_timings)),
                    "db_avg={0:.3f}".format(mean(query_timings)),
                ],
            )

        response[self._header] = " ".join(stats)
